"""Default port numbers for the URI schemes supported by urlparse."""

DEFAULT_PORTS = {
    'ftp': 21,
    'gopher': 70,
    'hdl': 2641,
    'http': 80,
    'https': 443,
    'imap': 143,
    'mms': 651,
    'news': 2009,
    'nntp': 119,
    'prospero': 191,
    'rsync': 873,
    'rtsp': 554,
    'rtspu': 554,
    'sftp': 115,
    'shttp': 80,
    'sip': 5060,
    'sips': 5061,
    'snews': 2009,
    'svn': 3690,
    'svn+ssh': 22,
    'telnet': 23,
}
